@@ -10,6 +10,8 @@
 #include <stdarg.h>
 #include <string>
 
+#include "include/buffer.h"
+
 namespace ceph {
 
 
@@ -25,6 +27,11 @@ class Formatter {
   virtual ~Formatter();
 
   virtual void flush(std::ostream& os) = 0;
+  void flush(bufferlist &bl) {
+    std::stringstream os;
+    flush(os);
+    bl.append(os.str());
+  }
   virtual void reset() = 0;
 
   virtual void open_array_section(const char *name) = 0;
@@ -38,6 +45,7 @@ class Formatter {
   virtual void dump_string(const char *name, std::string s) = 0;
   virtual std::ostream& dump_stream(const char *name) = 0;
   virtual void dump_format(const char *name, const char *fmt, ...) = 0;
+  virtual void dump_format_unquoted(const char *name, const char *fmt, ...) = 0;
   virtual int get_len() const = 0;
   virtual void write_raw_data(const char *data) = 0;
 
@@ -53,7 +61,7 @@ class Formatter {
   }
 };
 
-Formatter *new_formatter(const std::string &type);
+Formatter *new_formatter(const std::string type);
 
 class JSONFormatter : public Formatter {
  public:
@@ -72,6 +80,7 @@ class JSONFormatter : public Formatter {
   void dump_string(const char *name, std::string s);
   std::ostream& dump_stream(const char *name);
   void dump_format(const char *name, const char *fmt, ...);
+  void dump_format_unquoted(const char *name, const char *fmt, ...);
   int get_len() const;
   void write_raw_data(const char *data);
 
@@ -112,6 +121,7 @@ class XMLFormatter : public Formatter {
   void dump_string(const char *name, std::string s);
   std::ostream& dump_stream(const char *name);
   void dump_format(const char *name, const char *fmt, ...);
+  void dump_format_unquoted(const char *name, const char *fmt, ...);
   int get_len() const;
   void write_raw_data(const char *data);
 
