@@ -49,6 +49,17 @@ ThreadPool::ThreadPool(CephContext *cct_, string nm, int n, const char *option)
   }
 }
 
+void ThreadPool::TPHandle::suspend_tp_timeout()
+{
+  cct->get_heartbeat_map()->clear_timeout(hb);
+}
+
+void ThreadPool::TPHandle::reset_tp_timeout()
+{
+  cct->get_heartbeat_map()->reset_timeout(
+    hb, grace, suicide_grace);
+}
+
 ThreadPool::~ThreadPool()
 {
   assert(_threads.empty());
@@ -94,7 +105,7 @@ void ThreadPool::worker(WorkThread *wt)
       break;
     }
 
-    if (!_pause && work_queues.size()) {
+    if (!_pause && !work_queues.empty()) {
       WorkQueue_* wq;
       int tries = work_queues.size();
       bool did = false;
@@ -108,9 +119,10 @@ void ThreadPool::worker(WorkThread *wt)
 	  processing++;
 	  ldout(cct,12) << "worker wq " << wq->name << " start processing " << item
 			<< " (" << processing << " active)" << dendl;
-	  cct->get_heartbeat_map()->reset_timeout(hb, wq->timeout_interval, wq->suicide_interval);
+	  TPHandle tp_handle(cct, hb, wq->timeout_interval, wq->suicide_interval);
+	  tp_handle.reset_tp_timeout();
 	  _lock.Unlock();
-	  wq->_void_process(item);
+	  wq->_void_process(item, tp_handle);
 	  _lock.Lock();
 	  wq->_void_process_finish(item);
 	  processing--;
@@ -190,7 +202,7 @@ void ThreadPool::stop(bool clear_after)
   _lock.Unlock();
   for (set<WorkThread*>::iterator p = _threads.begin();
        p != _threads.end();
-       p++) {
+       ++p) {
     (*p)->join();
     delete *p;
   }
