@@ -67,6 +67,7 @@ enum {
   l_os_commit_len,
   l_os_commit_lat,
   l_os_j_full,
+  l_os_queue_lat,
   l_os_last,
 };
 
@@ -178,14 +179,23 @@ public:
       tolerate_collection_add_enoent = true;
     }
     void register_on_applied(Context *c) {
+      if (!c) return;
       on_applied.push_back(c);
     }
     void register_on_commit(Context *c) {
+      if (!c) return;
       on_commit.push_back(c);
     }
     void register_on_applied_sync(Context *c) {
+      if (!c) return;
       on_applied_sync.push_back(c);
     }
+    void register_on_complete(Context *c) {
+      if (!c) return;
+      RunOnDeleteRef _complete(new RunOnDelete(c));
+      register_on_applied(new ContainerContext<RunOnDeleteRef>(_complete));
+      register_on_commit(new ContainerContext<RunOnDeleteRef>(_complete));
+    }
 
     static void collect_contexts(
       list<Transaction *> &t,
