@@ -561,19 +561,26 @@ void ObjectCacher::perf_stop()
 
 /* private */
 ObjectCacher::Object *ObjectCacher::get_object(sobject_t oid, ObjectSet *oset,
-                                 object_locator_t &l)
+					       object_locator_t &l,
+					       uint64_t truncate_size,
+					       uint64_t truncate_seq)
 {
+  // XXX: Add handling of nspace in object_locator_t in cache
   assert(lock.is_locked());
   // have it?
   if ((uint32_t)l.pool < objects.size()) {
-    if (objects[l.pool].count(oid))
-      return objects[l.pool][oid];
+    if (objects[l.pool].count(oid)) {
+      Object *o = objects[l.pool][oid];
+      o->truncate_size = truncate_size;
+      o->truncate_seq = truncate_seq;
+      return o;
+    }
   } else {
     objects.resize(l.pool+1);
   }
 
   // create it.
-  Object *o = new Object(this, oid, oset, l);
+  Object *o = new Object(this, oid, oset, l, truncate_size, truncate_seq);
   objects[l.pool][oid] = o;
   ob_lru.lru_insert_top(o);
   return o;
@@ -607,13 +614,10 @@ void ObjectCacher::bh_read(BufferHead *bh)
   // finisher
   C_ReadFinish *onfinish = new C_ReadFinish(this, bh->ob, bh->last_read_tid,
 					    bh->start(), bh->length());
-
-  ObjectSet *oset = bh->ob->oset;
-
   // go
   writeback_handler.read(bh->ob->get_oid(), bh->ob->get_oloc(),
 			 bh->start(), bh->length(), bh->ob->get_snap(),
-			 &onfinish->bl, oset->truncate_size, oset->truncate_seq,
+			 &onfinish->bl, bh->ob->truncate_size, bh->ob->truncate_seq,
 			 onfinish);
 
   ++reads_outstanding;
@@ -803,14 +807,11 @@ void ObjectCacher::bh_write(BufferHead *bh)
   // finishers
   C_WriteCommit *oncommit = new C_WriteCommit(this, bh->ob->oloc.pool,
                                               bh->ob->get_soid(), bh->start(), bh->length());
-
-  ObjectSet *oset = bh->ob->oset;
-
   // go
   tid_t tid = writeback_handler.write(bh->ob->get_oid(), bh->ob->get_oloc(),
 				      bh->start(), bh->length(),
 				      bh->snapc, bh->bl, bh->last_write,
-				      oset->truncate_size, oset->truncate_seq,
+				      bh->ob->truncate_size, bh->ob->truncate_seq,
 				      oncommit);
   ldout(cct, 20) << " tid " << tid << " on " << bh->ob->get_oid() << dendl;
 
@@ -898,10 +899,11 @@ void ObjectCacher::bh_write_commit(int64_t poolid, sobject_t oid, loff_t start,
     ob->last_commit_tid = tid;
 
     // waiters?
-    list<Context*> ls;
     if (ob->waitfor_commit.count(tid)) {
+      list<Context*> ls;
       ls.splice(ls.begin(), ob->waitfor_commit[tid]);
       ob->waitfor_commit.erase(tid);
+      finish_contexts(cct, ls, r);
     }
 
     // is the entire object set now clean and fully committed?
@@ -913,9 +915,6 @@ void ObjectCacher::bh_write_commit(int64_t poolid, sobject_t oid, loff_t start,
 	oset->dirty_or_tx == 0) {        // nothing dirty/tx
       flush_set_callback(flush_set_callback_arg, oset);      
     }
-
-    if (!ls.empty())
-      finish_contexts(cct, ls, r);
   }
 }
 
@@ -1041,7 +1040,7 @@ int ObjectCacher::_readx(OSDRead *rd, ObjectSet *oset, Context *onfinish,
 
     // get Object cache
     sobject_t soid(ex_it->oid, rd->snap);
-    Object *o = get_object(soid, oset, ex_it->oloc);
+    Object *o = get_object(soid, oset, ex_it->oloc, ex_it->truncate_size, oset->truncate_seq);
     touch_ob(o);
 
     // does not exist and no hits?
@@ -1275,7 +1274,7 @@ int ObjectCacher::writex(OSDWrite *wr, ObjectSet *oset, Mutex& wait_on_lock,
        ++ex_it) {
     // get object cache
     sobject_t soid(ex_it->oid, CEPH_NOSNAP);
-    Object *o = get_object(soid, oset, ex_it->oloc);
+    Object *o = get_object(soid, oset, ex_it->oloc, ex_it->truncate_size, oset->truncate_seq);
 
     // map it all into a single bufferhead.
     BufferHead *bh = o->map_write(wr);
