@@ -3,13 +3,13 @@
 #define TIME_BUF_SIZE 128
 
 #include "rgw_op.h"
-#include "rgw_html_errors.h"
+#include "rgw_http_errors.h"
 #include "rgw_acl_s3.h"
 #include "rgw_policy_s3.h"
 
 #define RGW_AUTH_GRACE_MINS 15
 
-void rgw_get_errno_s3(struct rgw_html_errors *e, int err_no);
+void rgw_get_errno_s3(struct rgw_http_errors *e, int err_no);
 
 class RGWGetObj_ObjStore_S3 : public RGWGetObj_ObjStore
 {
@@ -143,12 +143,14 @@ public:
 };
 
 class RGWCopyObj_ObjStore_S3 : public RGWCopyObj_ObjStore {
+  bool sent_header;
 public:
-  RGWCopyObj_ObjStore_S3() {}
+  RGWCopyObj_ObjStore_S3() : sent_header(false) {}
   ~RGWCopyObj_ObjStore_S3() {}
 
   int init_dest_policy();
   int get_params();
+  void send_partial_response(off_t ofs);
   void send_response();
 };
 
@@ -182,6 +184,7 @@ public:
   RGWPutCORS_ObjStore_S3() {}
   ~RGWPutCORS_ObjStore_S3() {}
 
+  int get_params();
   void send_response();
 };
 
@@ -306,10 +309,10 @@ public:
 class RGWHandler_ObjStore_Bucket_S3 : public RGWHandler_ObjStore_S3 {
 protected:
   bool is_acl_op() {
-    return s->args.exists("acl");
+    return s->info.args.exists("acl");
   }
   bool is_cors_op() {
-      return s->args.exists("cors");
+      return s->info.args.exists("cors");
   }
   bool is_obj_update_op() {
     return is_acl_op() || is_cors_op();
@@ -330,10 +333,10 @@ public:
 class RGWHandler_ObjStore_Obj_S3 : public RGWHandler_ObjStore_S3 {
 protected:
   bool is_acl_op() {
-    return s->args.exists("acl");
+    return s->info.args.exists("acl");
   }
   bool is_cors_op() {
-      return s->args.exists("cors");
+      return s->info.args.exists("cors");
   }
   bool is_obj_update_op() {
     return is_acl_op();
